from .processor import TapProcessor
from ..Helpers import get_selected_streams


class LoanAccountsProcessor(TapProcessor):
    def _process_child_records(self, record):
        from .. import sync_endpoint_refactor

        total_records = super()._process_child_records(record)
        children_configs = self.generators[0].endpoint_config.get('children', {})
        for child_config_key in children_configs:
            if child_config_key in get_selected_streams(self.catalog):
                child_config = children_configs[child_config_key]
                child_config = dict(child_config)
                child_config['path'] = child_config['path'].format(str(record['id']))
                total_records += sync_endpoint_refactor(client=self.generators[0].client,
                                                        catalog=self.catalog,
                                                        state=self.generators[0].state,
                                                        stream_name=child_config_key,
                                                        sub_type='self',
                                                        config=self.generators[0].config,
                                                        endpoint_config=child_config)
        return total_records



from abc import ABC
from typing import List
from singer import utils

from ..Helpers import write_bookmark, transform_json, get_bookmark


class TapGenerator(ABC):
    def __init__(self, stream_name, client, config, state, sub_type, endpoint_config=None):
        self.stream_name = stream_name
        self.client = client
        self.config = config
        self.state = state
        self.sub_type = sub_type
        self.endpoint_config = endpoint_config
        self._init_config()
        self._init_endpoint_config()
        self._init_buffers()
        self._init_bookmarks()
        self._init_params()

    def _init_config(self):
        self.start_date = self.config.get('start_date')

    def _init_endpoint_config(self):
        if self.endpoint_config is None:
            self.endpoint_config = {}

    def _init_buffers(self):
        self.buffer: List = list()

    def _init_bookmarks(self):
        self.bookmark_query_field = self.endpoint_config.get('bookmark_query_field')
        self.bookmark_type = self.endpoint_config.get('bookmark_type')
        self.bookmark_field = self.endpoint_config.get('bookmark_field')
        if self.bookmark_type == "integer":
            self.last_bookmark_value = get_bookmark(self.state, self.stream_name, self.sub_type, 0)
        else:
            self.last_bookmark_value = get_bookmark(self.state, self.stream_name, self.sub_type, self.start_date)
        self.max_bookmark_value = self.last_bookmark_value

    def _init_params(self):
        self.time_extracted = None
        self.static_params = self.endpoint_config.get('params', {})
        self.offset = 0
        self.limit = self.client.page_size
        self.params = self.static_params

    def __all_fetch_batch_steps(self):
        self.prepare_batch()
        raw_batch = self.fetch_batch()
        self.buffer = self.transform_batch(raw_batch)
        self.last_batch_size = len(self.buffer)

    def __iter__(self):
        self.__all_fetch_batch_steps()
        return self

    def __next__(self):
        if not self.buffer:
            if self.last_batch_size < self.limit:
                raise StopIteration()
            self.offset += self.limit
            # self.write_bookmark()
            self.__all_fetch_batch_steps()
            if not self.buffer:
                raise StopIteration()
        return self.buffer.pop(0)

    def write_bookmark(self):
        if self.bookmark_field:
            write_bookmark(self.state,
                           self.stream_name,
                           self.sub_type,
                           self.max_bookmark_value)

    def prepare_batch(self):
        self.params = {
            "offset": self.offset,
            "limit": self.limit,
            **self.static_params
        }

    def fetch_batch(self):
        if self.bookmark_query_field:
            self.params[self.bookmark_query_field] = self.last_bookmark_value
        response = self.client.request(
            method=self.endpoint_config.get('api_method', 'GET'),
            path=self.endpoint_config.get('path'),
            version=self.endpoint_config.get('api_version', 'v2'),
            apikey_type=self.endpoint_config.get('apikey_type', None),
            params='&'.join([f'{key}={value}' for (key, value) in self.params.items()]),
            endpoint=self.stream_name,
            json=self.endpoint_config.get('body', None))
        self.time_extracted = utils.now()
        if isinstance(response, dict):
            return [response]
        return response

    def transform_batch(self, batch):
        data_key = self.endpoint_config.get('data_key', None)
        transformed_batch = list()
        if data_key is None:
            transformed_batch = transform_json(batch, self.stream_name)
        elif data_key in batch:
            transformed_batch = transform_json(batch, data_key)[data_key]
        return transformed_batch



from abc import ABC

from singer import write_record, Transformer, metadata, write_schema
from singer.utils import strptime_to_utc

from ..Helpers import transform_datetime, convert


class TapProcessor(ABC):
    def __init__(self, catalog, stream_name):
        self.generators = list()
        self.generator_values = dict()
        self.catalog = catalog
        self.stream_name = stream_name
        self.stream = self.catalog.get_stream(stream_name)
        self.schema = self.stream.schema.to_dict()
        self.stream_metadata = metadata.to_map(self.stream.metadata)
        self.deduplication_key = "id"

    def write_schema(self):
        stream = self.catalog.get_stream(self.stream_name)
        schema = stream.schema.to_dict()
        write_schema(self.stream_name, schema, stream.key_properties)

    def configure_processor_for_generators(self):
        id_fields = self.generators[0].endpoint_config['id_fields']
        if "id" not in id_fields:
            self.deduplication_key = id_fields[0]

    def process_streams_from_generators(self, generators):
        self.generators = generators
        self.configure_processor_for_generators()
        self.write_schema()
        record_count = 0
        for generator in self.generators:
            self.generator_values[iter(generator)] = None
        while True:
            # Populate list of values from generators (if any were removed)
            for iterator in list(self.generator_values):
                if self.generator_values[iterator] is None:
                    self.generator_values[iterator] = next(iterator, None)
                if self.generator_values[iterator] is None:
                    self.generator_values.pop(iterator)
            if not self.generator_values:
                break
            # Find lowest value in the list, and if two or more are equal, find max bookmark
            min_record_key = None
            min_record_value = None
            min_record_bookmark = None
            for iterator in self.generator_values:
                bookmark_field = convert(iterator.endpoint_config.get('bookmark_field', ''))
                # Different record
                if min_record_key is None \
                        or min_record_value > self.generator_values[iterator][self.deduplication_key]:
                    min_record_key = iterator
                    min_record_value = self.generator_values[iterator][self.deduplication_key]
                    if not bookmark_field:
                        continue
                    min_record_bookmark = self.generator_values[iterator][bookmark_field]
                # Same record
                elif min_record_value == self.generator_values[iterator][self.deduplication_key]:
                    if not bookmark_field:
                        continue

                    # Check the new bookmark against the min_record_key's bookmark
                    min_record_dttm = strptime_to_utc(min_record_bookmark)
                    record_dttm = strptime_to_utc(self.generator_values[iterator][bookmark_field])
                    # If min_record has bookmark smaller than record, then replace min_record (we want highest bookmark)
                    if min_record_dttm < record_dttm:
                        min_record_key = iterator
                        min_record_value = self.generator_values[iterator][self.deduplication_key]
                        min_record_bookmark = self.generator_values[iterator][bookmark_field]

            # Process the record
            record = self.generator_values[min_record_key]
            if self.process_record(record, min_record_key.time_extracted,
                                   min_record_key.endpoint_config.get('bookmark_field', '')):
                record_count += 1
                record_count += self._process_child_records(record)

            # Remove any record with the same deduplication_key from the list
            # (so we don't process the same record twice)
            for iterator in self.generator_values.keys():
                if min_record_value == self.generator_values[iterator][self.deduplication_key]:
                    self.generator_values[iterator] = None

        self.generators[0].write_bookmark()
        return record_count

    def _process_child_records(self, record):
        return 0

    def __is_record_past_bookmark(self, transformed_record, bookmark_field):
        is_record_past_bookmark = False
        bookmark_type = self.generators[0].endpoint_config.get('bookmark_type')
        bookmark_field = convert(bookmark_field)

        # Reset max_bookmark_value to new value if higher
        if bookmark_field and (bookmark_field in transformed_record):
            bookmark_dttm = strptime_to_utc(transformed_record[bookmark_field])
            if hasattr(self.generators[0], "max_bookmark_value"):
                max_bookmark_value_dttm = strptime_to_utc(self.generators[0].max_bookmark_value)
                if bookmark_dttm > max_bookmark_value_dttm:
                    self.generators[0].max_bookmark_value = transformed_record[bookmark_field]
            else:
                self.generators[0].max_bookmark_value = transformed_record[bookmark_field]

        if bookmark_field and (bookmark_field in transformed_record):
            if bookmark_type == 'integer':
                # Keep only records whose bookmark is after the last_integer
                if transformed_record[bookmark_field] >= self.generators[0].last_bookmark_value:
                    is_record_past_bookmark = True
            elif bookmark_type == 'datetime':
                last_dttm = transform_datetime(self.generators[0].last_bookmark_value)
                bookmark_dttm = transform_datetime(transformed_record[bookmark_field])
                # Keep only records whose bookmark is after the last_datetime
                if bookmark_dttm >= last_dttm:
                    is_record_past_bookmark = True
        else:
            is_record_past_bookmark = True

        return is_record_past_bookmark

    def process_record(self, record, time_extracted, bookmark_field):
        with Transformer() as transformer:
            transformed_record = transformer.transform(record,
                                                       self.schema,
                                                       self.stream_metadata)

        if self.__is_record_past_bookmark(transformed_record, bookmark_field):
            write_record(self.stream_name,
                         transformed_record,
                         time_extracted=time_extracted)
            return True
        return False



import singer

from .TapGenerators.loan_accounts_generator import LoanAccountsADGenerator, LoanAccountsLMGenerator
from .TapGenerators.loan_repayments_generator import LoanRepaymentsGenerator
from .TapProcessors.loan_accounts_processor import LoanAccountsProcessor
from .TapProcessors.loan_repayments_processor import LoanRepaymentsProcessor

LOGGER = singer.get_logger()

stream_generator_processor_dict = {
    "loan_accounts": ((LoanAccountsLMGenerator, LoanAccountsADGenerator), LoanAccountsProcessor),
    "loan_repayments": ((LoanRepaymentsGenerator,), LoanRepaymentsProcessor)
}


def sync_endpoint_refactor(client, catalog, state,
                           stream_name, sub_type, config, endpoint_config=None):
    generator_classes, processor_class = stream_generator_processor_dict[stream_name]
    generators = [generator_class(stream_name=stream_name,
                                  client=client,
                                  config=config,
                                  state=state,
                                  sub_type=sub_type,
                                  endpoint_config=endpoint_config)
                  for generator_class in generator_classes]
    processor = processor_class(catalog=catalog,
                                stream_name=stream_name)

    return processor.process_streams_from_generators(generators=generators)



import abc

from .generator import TapGenerator
from ..Helpers import transform_datetime, get_bookmark


class LoanAccountsGenerator(TapGenerator):
    @abc.abstractmethod
    def _init_endpoint_config(self):
        self.endpoint_config = {
            'path': 'loans:search',
            'api_version': 'v2',
            'api_method': 'POST',
            'params': {
                'detailsLevel': 'FULL',
                'paginationDetails': 'ON'
            },
            'body': {
                "sortingCriteria": {
                    "field": "id",
                    "order": "ASC"
                },
                "filterCriteria": [
                    {
                        "field": "",
                        "operator": "AFTER",
                        "value": transform_datetime(
                            get_bookmark(self.state, 'loan_accounts', 'self', self.start_date))[:10]
                    }
                ]
            },
            'bookmark_field': '',
            'bookmark_type': 'datetime',
            'id_fields': ['id'],
            'children': {
                'loan_repayments': {
                    'path': 'loans/{}/repayments',
                    'api_version': 'v1',
                    'api_method': 'GET',
                    'params': {
                        'detailsLevel': 'FULL',
                        'paginationDetails': 'ON'
                    },
                    'id_fields': ['encoded_key'],
                    'parent': 'loan_accounts'
                }
            }
        }


class LoanAccountsLMGenerator(LoanAccountsGenerator):
    def _init_endpoint_config(self):
        super()._init_endpoint_config()
        self.endpoint_config["body"]["filterCriteria"][0]["field"] = "lastModifiedDate"
        self.endpoint_config["bookmark_field"] = "lastModifiedDate"


class LoanAccountsADGenerator(LoanAccountsGenerator):
    def _init_endpoint_config(self):
        super()._init_endpoint_config()
        self.endpoint_config["body"]["filterCriteria"][0]["field"] = "lastAccountAppraisalDate"
        self.endpoint_config["bookmark_field"] = "lastAccountAppraisalDate"
